from itertools import product

from django_ltree.fields import PathValue


class PathGenerator(object):

    _alphabet = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'
    _default_label_size = 6  # Postgres limits this to 256

    def __init__(self, prefix=None, skip=None, label_size=None):
        self.skip_paths = [] if skip is None else skip[:]
        self.path_prefix = prefix if prefix else []
        self.product_iterator = product(self._alphabet, repeat=label_size or self._default_label_size)

    def __iter__(self):
        return self

    def __next__(self):
        for val in self.product_iterator:
            label = ''.join(val)
            path = PathValue(self.path_prefix + [label])
            if path not in self.skip_paths:
                return path

    next = __next__



from django import forms
from django.core.validators import RegexValidator
from django.db.models.fields import TextField
from django.forms.widgets import TextInput
from six import string_types


class PathValue(list):
    def __init__(self, val):  # type: (Union[list, str]) -> None
        if isinstance(val, str):
            val = val.split(".")
        elif isinstance(val, list):
            val = val
        else:
            raise ValueError("Invalid value for path: {}".format(val))

        super(PathValue, self).__init__(val)

    def __repr__(self):
        return str(self)

    def __str__(self):
        return ".".join(self)


class PathValueProxy:
    def __init__(self, field_name):  # type: (str) -> None
        self.field_name = field_name

    def __get__(self, instance, owner):
        if instance is None:
            return self

        value = instance.__dict__[self.field_name]

        if value is None:
            return value

        return PathValue(instance.__dict__[self.field_name])

    def __set__(self, instance, value):
        if instance is None:
            return self

        instance.__dict__[self.field_name] = value


path_label_validator = RegexValidator(
    r"^[A-Za-z0-9_.]+$",
    "A label is a sequence of alphanumeric characters and underscores separated by dots.",
    "invalid",
)


class PathFormField(forms.CharField):
    default_validators = [path_label_validator]


class PathField(TextField):
    default_validators = [path_label_validator]

    def db_type(self, connection):
        return "ltree"

    def formfield(self, **kwargs):
        kwargs["form_class"] = PathFormField
        kwargs["widget"] = TextInput(attrs={"class": "vTextField"})
        return super(PathField, self).formfield(**kwargs)

    def contribute_to_class(self, cls, name, private_only=False):
        super(PathField, self).contribute_to_class(cls, name)
        setattr(cls, self.name, PathValueProxy(self.name))

    def from_db_value(self, value, expression, connection, *args):
        if value is None:
            return value
        return PathValue(value)

    def get_prep_value(self, value):
        if value is None:
            return value
        return str(PathValue(value))

    def to_python(self, value):
        if value is None:
            return value
        elif isinstance(value, PathValue):
            return value

        return PathValue(value)

    def get_db_prep_value(self, value, connection, prepared=False):
        if value is None:
            return value
        elif isinstance(value, PathValue):
            return str(value)
        elif isinstance(value, (list, string_types)):
            return str(PathValue(value))

        raise ValueError("Unknown value type {}".format(type(value)))



import importlib

from django.apps import AppConfig


class DjangoLtreeConfig(AppConfig):
    name = "django_ltree"

    def ready(self):
        importlib.import_module("django_ltree.checks")
        importlib.import_module("django_ltree.lookups")
        importlib.import_module("django_ltree.functions")



from django.db import models

from .managers import TreeManager
from .fields import PathField, PathValue


class TreeModel(models.Model):
    path = PathField(unique=True)
    objects = TreeManager()

    class Meta:
        abstract = True
        ordering = ("path",)

    def label(self):
        return self.path[-1]

    def get_ancestors_paths(self):  # type: () -> List[List[str]]
        return [
            PathValue(self.path[:n]) for n, p in enumerate(self.path) if n > 0
        ]

    def ancestors(self):
        return (
            type(self)
            ._default_manager.filter(path__ancestors=self.path)
            .exclude(id=self.id)
        )

    def descendants(self):
        return (
            type(self)
            ._default_manager.filter(path__descendants=self.path)
            .exclude(id=self.id)
        )

    def parent(self):
        if len(self.path) > 1:
            return self.ancestors().exclude(id=self.id).last()

    def children(self):
        return self.descendants().filter(path__depth=len(self.path) + 1)

    def siblings(self):
        parent = self.path[:-1]
        return (
            type(self)
            ._default_manager.filter(path__descendants=".".join(parent))
            .filter(path__depth=len(self.path))
            .exclude(path=self.path)
        )



from django.db.models import IntegerField, Lookup, Transform

from .fields import PathField


class SimpleLookup(Lookup):
    lookup_operator = "="  # type: str

    def as_sql(self, compiler, connection):
        lhs, lhs_params = self.process_lhs(compiler, connection)
        rhs, rhs_params = self.process_rhs(compiler, connection)
        params = lhs_params + rhs_params
        return "{} {} {}".format(lhs, self.lookup_operator, rhs), params


@PathField.register_lookup
class EqualLookup(Lookup):
    lookup_name = "exact"

    def as_sql(self, compiler, connection):
        lhs, lhs_params = self.process_lhs(compiler, connection)
        rhs, rhs_params = self.process_rhs(compiler, connection)
        params = lhs_params + rhs_params

        return "{} = {}".format(lhs, rhs), params


@PathField.register_lookup
class AncestorLookup(SimpleLookup):
    lookup_name = "ancestors"
    lookup_operator = "@>"


@PathField.register_lookup
class DescendantLookup(SimpleLookup):
    lookup_name = "descendants"
    lookup_operator = "<@"


@PathField.register_lookup
class MatchLookup(SimpleLookup):
    lookup_name = "match"
    lookup_operator = "~"



from django.core.checks import Warning, register


@register
def check_database_backend_is_postgres(app_configs, **kwargs):
    from django.conf import settings
    errors = []
    valid_dbs=['postgres', 'postgis']
    if not any(d in settings.DATABASES['default']['ENGINE'] for d in valid_dbs):
        errors.append(Warning(
            'django_ltree needs postgres to support install the ltree extension.',
            hint='Use the postgres engine or ignore if you already use a custom engine for postgres'))

    return errors




# -*- coding: UTF-8 -*-
from setuptools import setup, find_packages


with open("README.md", "r") as fh:
    long_description = fh.read()


setup(
    name="django_ltree_fork",
    version="0.4.5",
    python_requires=">=2.7",
    url="https://github.com/boryszef/django-ltree",
    author="Borys Szefczyk",
    author_email="boryszef@gmail.com",
    description="django implemention of the ltree postgres extension - this is a backport to Python 2.7/Django 1.11",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="MIT",
    packages=find_packages(exclude=("example",)),
    install_requires=[
        "django>=1.11",
        "six"
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
    ],
    project_urls={
        "Source": "https://github.com/boryszef/django_ltree",
        "Tracker": "https://github.com/boryszef/django_ltree/issues",
    },
)
